use serde::{Deserialize, Serialize};

use crate::{Either, identifier::Identifier, item_stack::ItemComponent};

#[derive(Debug, Serialize, Deserialize, Clone, Eq, PartialEq)]
pub struct TextComponent {
    // Content
    #[serde(flatten)]
    pub content: TextComponentContent,
    // Children
    #[serde(default)]
    pub extra: Vec<TextComponent>,
    // Formatting
    #[serde(flatten)]
    pub style: TextComponentStyle,
    // Interactivity
    pub insertion: Option<String>,
    #[serde(rename = "clickEvent")]
    pub click_event: Option<ClickEvent>,
}

#[derive(Debug, Serialize, Deserialize, Clone, Eq, PartialEq)]
pub struct TextComponentStyle {
    pub color: Option<String>,
    pub font: Option<Identifier>,
    pub bold: Option<bool>,
    pub italic: Option<bool>,
    pub underlined: Option<bool>,
    pub strikethrough: Option<bool>,
    pub obfuscated: Option<bool>,
}

#[derive(Debug, Serialize, Deserialize, Clone, Eq, PartialEq)]
#[serde(tag = "action", content = "value", rename_all = "snake_case")]
pub enum ClickEvent {
    OpenUrl(String),
    OpenFile(String),
    RunCommand(String),
    SuggestCommand(String),
    ChangePage(String),
    CopyToClipboard(String),
}

#[derive(Debug, Serialize, Deserialize, Clone, Eq, PartialEq)]
#[serde(tag = "action", content = "contents", rename_all = "snake_case")]
pub enum HoverEvent {
    ShowText(Box<TextComponent>),
    ShowItem {
        id: Identifier,
        count: Option<i32>,
        #[serde(default)]
        components: Vec<ItemComponent>,
    },
    ShowEntity {
        // NOTE: Not sure if this is a TextComponent or just a string.
        name: String,
        #[serde(rename = "type")]
        ty: Identifier,
        id: Either<String, (u32, u32, u32, u32)>,
    },
}

#[derive(Debug, Serialize, Deserialize, Clone, Eq, PartialEq)]
#[serde(untagged)]
pub enum TextComponentContent {
    Text(TextComponentText),
    Translatable(TextComponentTranslatable),
    Score(TextComponentScore),
    Selector(TextComponentSelector),
    Keybind(TextComponentKeybind),
    Nbt(TextComponentNbt),
}

#[derive(Debug, Serialize, Deserialize, Clone, Eq, PartialEq)]
#[serde(tag = "type", rename = "text")]
pub struct TextComponentText {
    pub text: String,
}

#[derive(Debug, Serialize, Deserialize, Clone, Eq, PartialEq)]
#[serde(tag = "type", rename = "translatable")]
pub struct TextComponentTranslatable {
    pub translate: String,
    pub fallback: Option<String>,
    pub with: Option<Vec<TextComponent>>,
}

#[derive(Debug, Serialize, Deserialize, Clone, Eq, PartialEq)]
#[serde(tag = "type", rename = "score")]
pub struct TextComponentScore {
    pub score: TextComponentScoreInner,
}

#[derive(Debug, Serialize, Deserialize, Clone, Eq, PartialEq)]
pub struct TextComponentScoreInner {
    pub name: String,
    pub objective: String,
}

#[derive(Debug, Serialize, Deserialize, Clone, Eq, PartialEq)]
#[serde(tag = "type", rename = "selector")]
pub struct TextComponentSelector {
    pub selector: String,
    // TODO: Instead of an option, make this default to the default value.
    pub separator: Option<Box<TextComponent>>,
}

#[derive(Debug, Serialize, Deserialize, Clone, Eq, PartialEq)]
#[serde(tag = "type", rename = "keybind")]
pub struct TextComponentKeybind {
    pub keybind: String, // TODO: Maybe restrict this to the allowed keybinds, or maybe leave that
                         // for the client.
}

#[derive(Debug, Serialize, Deserialize, Clone, Eq, PartialEq)]
#[serde(tag = "type", rename = "nbt")]
pub struct TextComponentNbt {
    pub source: String,
    pub nbt: String,
    #[serde(default)]
    pub interpret: bool,
    // TODO: Instead of an option, make this default to the default value.
    pub separator: Option<Box<TextComponent>>,
    pub block: Option<String>,
    pub entity: Option<String>,
    pub storage: Option<String>,
}

#[cfg(test)]
mod test {
    use crate::text_component::*;

    #[test]
    fn test_text_component() {
        let text = r#"{"text":"Hello, world!"}"#;
        let text_component: TextComponent = serde_json::from_str(text).unwrap();
        assert_eq!(
            text_component,
            TextComponent {
                content: TextComponentContent::Text(TextComponentText {
                    text: "Hello, world!".to_string(),
                }),
            }
        );
    }

    #[test]
    fn test_translatable_text_component() {
        let text = r#"{"translate":"chat.type.text","with":[{"text":"Hello, world!"}]}"#;
        let text_component: TextComponent = serde_json::from_str(text).unwrap();
        assert_eq!(
            text_component,
            TextComponent {
                content: TextComponentContent::Translatable(TextComponentTranslatable {
                    translate: "chat.type.text".to_string(),
                    fallback: None,
                    with: Some(vec![TextComponent {
                        content: TextComponentContent::Text(TextComponentText {
                            text: "Hello, world!".to_string(),
                        }),
                    }]),
                }),
            }
        );
    }
}
