use bytes::{Buf, BufMut};
use tokio::{io::AsyncWriteExt, net::tcp::OwnedWriteHalf};

use crate::packet_encodable::{PacketDecodeError, PacketEncodable};

const SEGMENT_BITS: u8 = 0x7F;
const CONTINUE_BIT: u8 = 0x80;

#[derive(Debug, Clone, Copy)]
pub struct VarInt(i32);

impl VarInt {
    pub fn encoded_len(&self) -> usize {
        match self.0 {
            0 => 1,
            n => (31 - n.leading_zeros() as usize) / 7 + 1,
        }
    }

    pub async fn write_to_stream(&self, stream: &mut OwnedWriteHalf) -> Result<(), std::io::Error> {
        let mut value = self.0;

        loop {
            if value & !(SEGMENT_BITS as i32) == 0 {
                stream.write_u8(value as u8).await?;
                break;
            }

            stream
                .write_u8(((value & (SEGMENT_BITS as i32)) | (CONTINUE_BIT as i32)) as u8)
                .await?;
            value >>= 7;
        }

        Ok(())
    }
}

impl PacketEncodable for VarInt {
    fn encode_packet(
        &self,
        buf: &mut impl BufMut,
    ) -> Result<(), crate::packet_encodable::PacketEncodeError> {
        let mut value = self.0;

        loop {
            if value & !(SEGMENT_BITS as i32) == 0 {
                buf.put_u8(value as u8);
                break;
            }

            buf.put_u8(((value & (SEGMENT_BITS as i32)) | (CONTINUE_BIT as i32)) as u8);
            value >>= 7;
        }

        Ok(())
    }

    fn decode_packet(buf: &mut impl Buf) -> Result<Self, PacketDecodeError> {
        let mut value: i32 = 0;
        let mut pos = 0;

        loop {
            if !buf.has_remaining() {
                // The data is incomplete
                return Err(PacketDecodeError::Incomplete);
            }

            let current_byte = buf.get_u8();

            value |= ((current_byte & SEGMENT_BITS) as i32) << pos;
            if current_byte & 0x80 == 0 {
                break;
            }
            pos += 7;

            if pos > 32 {
                // The value is too large
                return Err(PacketDecodeError::InvalidVarInt);
            }
        }

        Ok(VarInt(value))
    }
}

impl From<VarInt> for i32 {
    fn from(val: VarInt) -> Self {
        val.0
    }
}

impl From<VarInt> for usize {
    fn from(val: VarInt) -> Self {
        val.0 as usize
    }
}

impl From<i32> for VarInt {
    fn from(value: i32) -> Self {
        VarInt(value)
    }
}

impl From<usize> for VarInt {
    fn from(value: usize) -> Self {
        VarInt(value as i32)
    }
}
