use bytes::{Buf, BufMut};
use potato_data::identifier::Identifier;
use serde::{Serialize, de::DeserializeOwned};
use std::fmt;

use thiserror::Error;
use uuid::Uuid;

use crate::datatypes::var_int::VarInt;

pub trait PacketEncodable: Sized + fmt::Debug {
    fn encode_packet(&self, buf: &mut impl BufMut) -> Result<(), PacketEncodeError>;
    fn decode_packet(buf: &mut impl Buf) -> Result<Self, PacketDecodeError>;
}

#[derive(Error, Debug)]
pub enum PacketDecodeError {
    #[error("IO error while reading packet {0}")]
    IOError(#[from] std::io::Error),
    #[error("Invalid UTF-8 sequence in string")]
    FromUtf8Error(#[from] std::string::FromUtf8Error),
    #[error("Error while decoding JSON in packet {0}")]
    JsonError(#[from] serde_json::Error),
    #[error("Unexpected end of packet")]
    UnexpectedEndOfPacket,
    #[error("Unknown enum variant in enum {name} with value {value}")]
    UnkownEnumVariant { name: &'static str, value: i32 },
    #[error("Invalid identifier: {0}")]
    InvalidIdentifier(String),
    #[error("Error while decoding NBT in packet {0}")]
    NbtError(#[from] fastnbt::error::Error),
    #[error("Not enough data to decode the packet yet")]
    Incomplete,
    #[error("Invalid VarInt")]
    InvalidVarInt,
}

#[derive(Error, Debug)]
pub enum PacketEncodeError {
    #[error("IO error while writing packet")]
    IOError(#[from] std::io::Error),
    #[error("Error while encoding JSON in packet {0}")]
    JsonError(#[from] serde_json::Error),
    #[error("Invalid packet id: {0}")]
    InvalidPacketId(i32),
    #[error("Error while encoding NBT in packet {0}")]
    NbtError(#[from] fastnbt::error::Error),
}

impl PacketEncodable for Uuid {
    fn encode_packet(&self, buf: &mut impl BufMut) -> Result<(), PacketEncodeError> {
        buf.put_u128(self.as_u128());
        Ok(())
    }

    fn decode_packet(buf: &mut impl Buf) -> Result<Self, PacketDecodeError> {
        let data = buf.get_u128();
        Ok(Uuid::from_u128(data))
    }
}

#[derive(Debug)]
pub struct Json<T: Serialize + DeserializeOwned + fmt::Debug>(pub T);

impl<T> PacketEncodable for Json<T>
where
    T: Serialize + DeserializeOwned + fmt::Debug,
{
    fn encode_packet(&self, buf: &mut impl BufMut) -> Result<(), PacketEncodeError> {
        let json = serde_json::to_string(&self.0)?;
        json.encode_packet(buf)
    }

    fn decode_packet(buf: &mut impl Buf) -> Result<Self, PacketDecodeError> {
        let json = String::decode_packet(buf)?;
        Ok(Self(serde_json::from_str(&json)?))
    }
}

#[derive(Debug)]
pub struct Nbt<T>(pub T)
where
    T: Serialize + DeserializeOwned + fmt::Debug;

impl<T> PacketEncodable for Nbt<T>
where
    T: Serialize + DeserializeOwned + fmt::Debug,
{
    fn encode_packet(&self, buf: &mut impl BufMut) -> Result<(), PacketEncodeError> {
        fastnbt::to_writer_with_opts(buf.writer(), &self.0, fastnbt::SerOpts::network_nbt())?;
        Ok(())
    }

    fn decode_packet(buf: &mut impl Buf) -> Result<Self, PacketDecodeError> {
        todo!("Decode NBT in packets")
        // let value = fastnbt::from_bytes(buf)?;
        //
        // Ok(Nbt(value))
    }
}

impl<T> PacketEncodable for Option<T>
where
    T: PacketEncodable,
{
    fn encode_packet(&self, buf: &mut impl BufMut) -> Result<(), PacketEncodeError> {
        let present = self.is_some();
        present.encode_packet(buf)?;
        if let Some(value) = self {
            value.encode_packet(buf)?;
        }

        Ok(())
    }

    fn decode_packet(buf: &mut impl Buf) -> Result<Self, PacketDecodeError> {
        let present = bool::decode_packet(buf)?;

        if present {
            Ok(Some(T::decode_packet(buf)?))
        } else {
            Ok(None)
        }
    }
}

impl<T> PacketEncodable for Vec<T>
where
    T: PacketEncodable,
{
    fn encode_packet(&self, buf: &mut impl BufMut) -> Result<(), PacketEncodeError> {
        let length: VarInt = self.len().into();
        length.encode_packet(buf)?;

        for item in self {
            item.encode_packet(buf)?;
        }

        Ok(())
    }

    fn decode_packet(buf: &mut impl Buf) -> Result<Self, PacketDecodeError> {
        let length: usize = VarInt::decode_packet(buf)?.into();
        let mut vec = Vec::with_capacity(length);
        for _ in 0..length {
            vec.push(T::decode_packet(buf)?);
        }

        Ok(vec)
    }
}

impl PacketEncodable for String {
    fn encode_packet(&self, buf: &mut impl BufMut) -> Result<(), PacketEncodeError> {
        let lenght: VarInt = self.len().into();
        lenght.encode_packet(buf)?;

        buf.put_slice(self.as_bytes());
        Ok(())
    }

    fn decode_packet(buf: &mut impl Buf) -> Result<Self, PacketDecodeError> {
        let lenght: usize = VarInt::decode_packet(buf)?.into();
        let mut buffer = vec![0; lenght];
        buf.copy_to_slice(&mut buffer);
        let value = String::from_utf8(buffer)?;

        Ok(value)
    }
}

impl PacketEncodable for bool {
    fn encode_packet(&self, buf: &mut impl BufMut) -> Result<(), PacketEncodeError> {
        buf.put_u8(if *self { 1 } else { 0 });
        Ok(())
    }

    fn decode_packet(buf: &mut impl Buf) -> Result<Self, PacketDecodeError> {
        let v = buf.get_u8();
        Ok(v != 0)
    }
}

impl PacketEncodable for Identifier {
    fn encode_packet(&self, buf: &mut impl BufMut) -> Result<(), PacketEncodeError> {
        let val: String = self.into();
        val.encode_packet(buf)
    }

    fn decode_packet(buf: &mut impl Buf) -> Result<Self, PacketDecodeError> {
        let str = String::decode_packet(buf)?;
        let mut parts = str.split(":");

        Ok(Self::new(
            parts
                .next()
                .ok_or_else(|| PacketDecodeError::InvalidIdentifier(str.clone()))?
                .to_string(),
            parts
                .next()
                .ok_or_else(|| PacketDecodeError::InvalidIdentifier(str.clone()))?
                .to_string(),
        ))
    }
}

macro_rules! number_impl {
    ($($type:ty, $read_fn:tt, $write_fn:tt),* $(,)?) => {
        $(
            impl PacketEncodable for $type {
                fn encode_packet(&self, buf: &mut impl BufMut) -> Result<(), PacketEncodeError> {
                    buf.$write_fn(*self);
                    Ok(())
                }

                fn decode_packet(buf: &mut impl Buf) -> Result<Self, PacketDecodeError> {
                    let value = buf.$read_fn();
                    Ok(value)
                }
            }
        )*
    }
}

number_impl! {
    u8, get_u8, put_u8,
    u16, get_u16, put_u16,
    u32, get_u32, put_u32,
    u64, get_u64, put_u64,

    i8, get_i8, put_i8,
    i16, get_i16, put_i16,
    i32, get_i32, put_i32,
    i64, get_i64, put_i64,

    f32, get_f32, put_f32,
    f64, get_f64, put_f64,
}
